@@ -1,10 +1,8 @@
 #include "jsi.h"
-#include "jscompile.h"
-#include "jsvalue.h"
-#include "jsrun.h"
-
 #include "utf.h"
 
+#include <assert.h>
+
 static void jsR_run(js_State *J, js_Function *F);
 
 /* Push values on stack */
@@ -13,9 +11,17 @@ static void jsR_run(js_State *J, js_Function *F);
 #define TOP (J->top)
 #define BOT (J->bot)
 
+static void js_trystackoverflow(js_State *J)
+{
+	STACK[TOP].t.type = JS_TLITSTR;
+	STACK[TOP].u.litstr = "exception stack overflow";
+	++TOP;
+	js_throw(J);
+}
+
 static void js_stackoverflow(js_State *J)
 {
-	STACK[TOP].type = JS_TLITSTR;
+	STACK[TOP].t.type = JS_TLITSTR;
 	STACK[TOP].u.litstr = "stack overflow";
 	++TOP;
 	js_throw(J);
@@ -23,15 +29,35 @@ static void js_stackoverflow(js_State *J)
 
 static void js_outofmemory(js_State *J)
 {
-	STACK[TOP].type = JS_TLITSTR;
+	STACK[TOP].t.type = JS_TLITSTR;
 	STACK[TOP].u.litstr = "out of memory";
 	++TOP;
 	js_throw(J);
 }
 
+static void js_runlimit(js_State *J)
+{
+	STACK[TOP].t.type = JS_TLITSTR;
+	STACK[TOP].u.litstr = "script ran too long";
+	++TOP;
+	js_throw(J);
+}
+
+void js_setlimit(js_State *J, int runlimit, int memlimit)
+{
+	J->runlimit = runlimit;
+	J->memlimit = memlimit;
+}
+
 void *js_malloc(js_State *J, int size)
 {
-	void *ptr = J->alloc(J->actx, NULL, size);
+	void *ptr;
+	if (J->memlimit > 0) {
+		if (size >= J->memlimit)
+			js_outofmemory(J);
+		J->memlimit -= size;
+	}
+	ptr = J->alloc(J->actx, NULL, size);
 	if (!ptr)
 		js_outofmemory(J);
 	return ptr;
@@ -39,6 +65,12 @@ void *js_malloc(js_State *J, int size)
 
 void *js_realloc(js_State *J, void *ptr, int size)
 {
+	if (J->memlimit > 0) {
+		// TODO: track released memory
+		if (size >= J->memlimit)
+			js_outofmemory(J);
+		J->memlimit -= size;
+	}
 	ptr = J->alloc(J->actx, ptr, size);
 	if (!ptr)
 		js_outofmemory(J);
@@ -55,6 +87,7 @@ char *js_strdup(js_State *J, const char *s)
 
 void js_free(js_State *J, void *ptr)
 {
+	// TODO: track released memory (J->memlimit)
 	J->alloc(J->actx, ptr, 0);
 }
 
@@ -82,21 +115,21 @@ void js_pushvalue(js_State *J, js_Value v)
 void js_pushundefined(js_State *J)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TUNDEFINED;
+	STACK[TOP].t.type = JS_TUNDEFINED;
 	++TOP;
 }
 
 void js_pushnull(js_State *J)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TNULL;
+	STACK[TOP].t.type = JS_TNULL;
 	++TOP;
 }
 
 void js_pushboolean(js_State *J, int v)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TBOOLEAN;
+	STACK[TOP].t.type = JS_TBOOLEAN;
 	STACK[TOP].u.boolean = !!v;
 	++TOP;
 }
@@ -104,22 +137,24 @@ void js_pushboolean(js_State *J, int v)
 void js_pushnumber(js_State *J, double v)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TNUMBER;
+	STACK[TOP].t.type = JS_TNUMBER;
 	STACK[TOP].u.number = v;
 	++TOP;
 }
 
 void js_pushstring(js_State *J, const char *v)
 {
-	int n = strlen(v);
+	size_t n = strlen(v);
+	if (n > JS_STRLIMIT)
+		js_rangeerror(J, "invalid string length");
 	CHECKSTACK(1);
-	if (n <= soffsetof(js_Value, type)) {
+	if (n <= soffsetof(js_Value, t.type)) {
 		char *s = STACK[TOP].u.shrstr;
 		while (n--) *s++ = *v++;
 		*s = 0;
-		STACK[TOP].type = JS_TSHRSTR;
+		STACK[TOP].t.type = JS_TSHRSTR;
 	} else {
-		STACK[TOP].type = JS_TMEMSTR;
+		STACK[TOP].t.type = JS_TMEMSTR;
 		STACK[TOP].u.memstr = jsV_newmemstring(J, v, n);
 	}
 	++TOP;
@@ -127,14 +162,16 @@ void js_pushstring(js_State *J, const char *v)
 
 void js_pushlstring(js_State *J, const char *v, int n)
 {
+	if (n > JS_STRLIMIT)
+		js_rangeerror(J, "invalid string length");
 	CHECKSTACK(1);
-	if (n <= soffsetof(js_Value, type)) {
+	if (n <= soffsetof(js_Value, t.type)) {
 		char *s = STACK[TOP].u.shrstr;
 		while (n--) *s++ = *v++;
 		*s = 0;
-		STACK[TOP].type = JS_TSHRSTR;
+		STACK[TOP].t.type = JS_TSHRSTR;
 	} else {
-		STACK[TOP].type = JS_TMEMSTR;
+		STACK[TOP].t.type = JS_TMEMSTR;
 		STACK[TOP].u.memstr = jsV_newmemstring(J, v, n);
 	}
 	++TOP;
@@ -143,7 +180,7 @@ void js_pushlstring(js_State *J, const char *v, int n)
 void js_pushliteral(js_State *J, const char *v)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TLITSTR;
+	STACK[TOP].t.type = JS_TLITSTR;
 	STACK[TOP].u.litstr = v;
 	++TOP;
 }
@@ -151,7 +188,7 @@ void js_pushliteral(js_State *J, const char *v)
 void js_pushobject(js_State *J, js_Object *v)
 {
 	CHECKSTACK(1);
-	STACK[TOP].type = JS_TOBJECT;
+	STACK[TOP].t.type = JS_TOBJECT;
 	STACK[TOP].u.object = v;
 	++TOP;
 }
@@ -164,15 +201,25 @@ void js_pushglobal(js_State *J)
 void js_currentfunction(js_State *J)
 {
 	CHECKSTACK(1);
-	STACK[TOP] = STACK[BOT-1];
+	if (BOT > 0)
+		STACK[TOP] = STACK[BOT-1];
+	else
+		STACK[TOP].t.type = JS_TUNDEFINED;
 	++TOP;
 }
 
+void *js_currentfunctiondata(js_State *J)
+{
+	if (BOT > 0)
+		return STACK[BOT-1].u.object->u.c.data;
+	return NULL;
+}
+
 /* Read values from stack */
 
 static js_Value *stackidx(js_State *J, int idx)
 {
-	static js_Value undefined = { {0}, {0}, JS_TUNDEFINED };
+	static js_Value undefined = { { {0}, JS_TUNDEFINED } };
 	idx = idx < 0 ? TOP + idx : BOT + idx;
 	if (idx < 0 || idx >= TOP)
 		return &undefined;
@@ -184,20 +231,20 @@ js_Value *js_tovalue(js_State *J, int idx)
 	return stackidx(J, idx);
 }
 
-int js_isdefined(js_State *J, int idx) { return stackidx(J, idx)->type != JS_TUNDEFINED; }
-int js_isundefined(js_State *J, int idx) { return stackidx(J, idx)->type == JS_TUNDEFINED; }
-int js_isnull(js_State *J, int idx) { return stackidx(J, idx)->type == JS_TNULL; }
-int js_isboolean(js_State *J, int idx) { return stackidx(J, idx)->type == JS_TBOOLEAN; }
-int js_isnumber(js_State *J, int idx) { return stackidx(J, idx)->type == JS_TNUMBER; }
-int js_isstring(js_State *J, int idx) { enum js_Type t = stackidx(J, idx)->type; return t == JS_TSHRSTR || t == JS_TLITSTR || t == JS_TMEMSTR; }
-int js_isprimitive(js_State *J, int idx) { return stackidx(J, idx)->type != JS_TOBJECT; }
-int js_isobject(js_State *J, int idx) { return stackidx(J, idx)->type == JS_TOBJECT; }
-int js_iscoercible(js_State *J, int idx) { js_Value *v = stackidx(J, idx); return v->type != JS_TUNDEFINED && v->type != JS_TNULL; }
+int js_isdefined(js_State *J, int idx) { return stackidx(J, idx)->t.type != JS_TUNDEFINED; }
+int js_isundefined(js_State *J, int idx) { return stackidx(J, idx)->t.type == JS_TUNDEFINED; }
+int js_isnull(js_State *J, int idx) { return stackidx(J, idx)->t.type == JS_TNULL; }
+int js_isboolean(js_State *J, int idx) { return stackidx(J, idx)->t.type == JS_TBOOLEAN; }
+int js_isnumber(js_State *J, int idx) { return stackidx(J, idx)->t.type == JS_TNUMBER; }
+int js_isstring(js_State *J, int idx) { enum js_Type t = stackidx(J, idx)->t.type; return t == JS_TSHRSTR || t == JS_TLITSTR || t == JS_TMEMSTR; }
+int js_isprimitive(js_State *J, int idx) { return stackidx(J, idx)->t.type != JS_TOBJECT; }
+int js_isobject(js_State *J, int idx) { return stackidx(J, idx)->t.type == JS_TOBJECT; }
+int js_iscoercible(js_State *J, int idx) { js_Value *v = stackidx(J, idx); return v->t.type != JS_TUNDEFINED && v->t.type != JS_TNULL; }
 
 int js_iscallable(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	if (v->type == JS_TOBJECT)
+	if (v->t.type == JS_TOBJECT)
 		return v->u.object->type == JS_CFUNCTION ||
 			v->u.object->type == JS_CSCRIPT ||
 			v->u.object->type == JS_CCFUNCTION;
@@ -207,19 +254,19 @@ int js_iscallable(js_State *J, int idx)
 int js_isarray(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	return v->type == JS_TOBJECT && v->u.object->type == JS_CARRAY;
+	return v->t.type == JS_TOBJECT && v->u.object->type == JS_CARRAY;
 }
 
 int js_isregexp(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	return v->type == JS_TOBJECT && v->u.object->type == JS_CREGEXP;
+	return v->t.type == JS_TOBJECT && v->u.object->type == JS_CREGEXP;
 }
 
 int js_isuserdata(js_State *J, int idx, const char *tag)
 {
 	js_Value *v = stackidx(J, idx);
-	if (v->type == JS_TOBJECT && v->u.object->type == JS_CUSERDATA)
+	if (v->t.type == JS_TOBJECT && v->u.object->type == JS_CUSERDATA)
 		return !strcmp(tag, v->u.object->u.user.tag);
 	return 0;
 }
@@ -227,13 +274,13 @@ int js_isuserdata(js_State *J, int idx, const char *tag)
 int js_iserror(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	return v->type == JS_TOBJECT && v->u.object->type == JS_CERROR;
+	return v->t.type == JS_TOBJECT && v->u.object->type == JS_CERROR;
 }
 
 const char *js_typeof(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	switch (v->type) {
+	switch (v->t.type) {
 	default:
 	case JS_TSHRSTR: return "string";
 	case JS_TUNDEFINED: return "undefined";
@@ -252,7 +299,7 @@ const char *js_typeof(js_State *J, int idx)
 int js_type(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	switch (v->type) {
+	switch (v->t.type) {
 	default:
 	case JS_TSHRSTR: return JS_ISSTRING;
 	case JS_TUNDEFINED: return JS_ISUNDEFINED;
@@ -321,7 +368,7 @@ void js_toprimitive(js_State *J, int idx, int hint)
 js_Regexp *js_toregexp(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	if (v->type == JS_TOBJECT && v->u.object->type == JS_CREGEXP)
+	if (v->t.type == JS_TOBJECT && v->u.object->type == JS_CREGEXP)
 		return &v->u.object->u.r;
 	js_typeerror(J, "not a regexp");
 }
@@ -329,7 +376,7 @@ js_Regexp *js_toregexp(js_State *J, int idx)
 void *js_touserdata(js_State *J, int idx, const char *tag)
 {
 	js_Value *v = stackidx(J, idx);
-	if (v->type == JS_TOBJECT && v->u.object->type == JS_CUSERDATA)
+	if (v->t.type == JS_TOBJECT && v->u.object->type == JS_CUSERDATA)
 		if (!strcmp(tag, v->u.object->u.user.tag))
 			return v->u.object->u.user.data;
 	js_typeerror(J, "not a %s", tag);
@@ -338,9 +385,9 @@ void *js_touserdata(js_State *J, int idx, const char *tag)
 static js_Object *jsR_tofunction(js_State *J, int idx)
 {
 	js_Value *v = stackidx(J, idx);
-	if (v->type == JS_TUNDEFINED || v->type == JS_TNULL)
+	if (v->t.type == JS_TUNDEFINED || v->t.type == JS_TNULL)
 		return NULL;
-	if (v->type == JS_TOBJECT)
+	if (v->t.type == JS_TOBJECT)
 		if (v->u.object->type == JS_CFUNCTION || v->u.object->type == JS_CCFUNCTION)
 			return v->u.object;
 	js_typeerror(J, "not a function");
@@ -495,6 +542,29 @@ static void js_pushrune(js_State *J, Rune rune)
 	}
 }
 
+void jsR_unflattenarray(js_State *J, js_Object *obj) {
+	if (obj->type == JS_CARRAY && obj->u.a.simple) {
+		js_Property *ref;
+		int i;
+		char name[32];
+		if (js_try(J)) {
+			obj->properties = NULL;
+			js_throw(J);
+		}
+		for (i = 0; i < obj->u.a.flat_length; ++i) {
+			js_itoa(name, i);
+			ref = jsV_setproperty(J, obj, name);
+			ref->value = obj->u.a.array[i];
+		}
+		js_free(J, obj->u.a.array);
+		obj->u.a.simple = 0;
+		obj->u.a.flat_length = 0;
+		obj->u.a.flat_capacity = 0;
+		obj->u.a.array = NULL;
+		js_endtry(J);
+	}
+}
+
 static int jsR_hasproperty(js_State *J, js_Object *obj, const char *name)
 {
 	js_Property *ref;
@@ -505,6 +575,15 @@ static int jsR_hasproperty(js_State *J, js_Object *obj, const char *name)
 			js_pushnumber(J, obj->u.a.length);
 			return 1;
 		}
+		if (obj->u.a.simple) {
+			if (js_isarrayindex(J, name, &k)) {
+				if (k >= 0 && k < obj->u.a.flat_length) {
+					js_pushvalue(J, obj->u.a.array[k]);
+					return 1;
+				}
+				return 0;
+			}
+		}
 	}
 
 	else if (obj->type == JS_CSTRING) {
@@ -522,7 +601,7 @@ static int jsR_hasproperty(js_State *J, js_Object *obj, const char *name)
 
 	else if (obj->type == JS_CREGEXP) {
 		if (!strcmp(name, "source")) {
-			js_pushliteral(J, obj->u.r.source);
+			js_pushstring(J, obj->u.r.source);
 			return 1;
 		}
 		if (!strcmp(name, "global")) {
@@ -569,6 +648,50 @@ static void jsR_getproperty(js_State *J, js_Object *obj, const char *name)
 		js_pushundefined(J);
 }
 
+static int jsR_hasindex(js_State *J, js_Object *obj, int k)
+{
+	char buf[32];
+	if (obj->type == JS_CARRAY && obj->u.a.simple) {
+		if (k >= 0 && k < obj->u.a.flat_length) {
+			js_pushvalue(J, obj->u.a.array[k]);
+			return 1;
+		}
+		return 0;
+	}
+	return jsR_hasproperty(J, obj, js_itoa(buf, k));
+}
+
+static void jsR_getindex(js_State *J, js_Object *obj, int k)
+{
+	if (!jsR_hasindex(J, obj, k))
+		js_pushundefined(J);
+}
+
+static void jsR_setarrayindex(js_State *J, js_Object *obj, int k, js_Value *value)
+{
+	int newlen = k + 1;
+	assert(obj->u.a.simple);
+	assert(k >= 0);
+	if (newlen > JS_ARRAYLIMIT)
+		js_rangeerror(J, "array too large");
+	if (newlen > obj->u.a.flat_length) {
+		assert(newlen == obj->u.a.flat_length + 1);
+		if (newlen > obj->u.a.flat_capacity) {
+			int newcap = obj->u.a.flat_capacity;
+			if (newcap == 0)
+				newcap = 8;
+			while (newcap < newlen)
+				newcap <<= 1;
+			obj->u.a.array = js_realloc(J, obj->u.a.array, newcap * sizeof(js_Value));
+			obj->u.a.flat_capacity = newcap;
+		}
+		obj->u.a.flat_length = newlen;
+	}
+	if (newlen > obj->u.a.length)
+		obj->u.a.length = newlen;
+	obj->u.a.array[k] = *value;
+}
+
 static void jsR_setproperty(js_State *J, js_Object *obj, const char *name, int transient)
 {
 	js_Value *value = stackidx(J, -1);
@@ -582,12 +705,32 @@ static void jsR_setproperty(js_State *J, js_Object *obj, const char *name, int t
 			int newlen = jsV_numbertointeger(rawlen);
 			if (newlen != rawlen || newlen < 0)
 				js_rangeerror(J, "invalid array length");
-			jsV_resizearray(J, obj, newlen);
+			if (newlen > JS_ARRAYLIMIT)
+				js_rangeerror(J, "array too large");
+			if (obj->u.a.simple) {
+				obj->u.a.length = newlen;
+				if (newlen <= obj->u.a.flat_length)
+					obj->u.a.flat_length = newlen;
+			} else  {
+				jsV_resizearray(J, obj, newlen);
+			}
 			return;
 		}
-		if (js_isarrayindex(J, name, &k))
-			if (k >= obj->u.a.length)
-				obj->u.a.length = k + 1;
+
+		if (js_isarrayindex(J, name, &k)) {
+			if (obj->u.a.simple) {
+				if (k >= 0 && k <= obj->u.a.flat_length) {
+					jsR_setarrayindex(J, obj, k, value);
+				} else {
+					jsR_unflattenarray(J, obj);
+					if (obj->u.a.length < k + 1)
+						obj->u.a.length = k + 1;
+				}
+			} else {
+				if (obj->u.a.length < k + 1)
+					obj->u.a.length = k + 1;
+			}
+		}
 	}
 
 	else if (obj->type == JS_CSTRING) {
@@ -657,6 +800,16 @@ readonly:
 		js_typeerror(J, "'%s' is read-only", name);
 }
 
+static void jsR_setindex(js_State *J, js_Object *obj, int k, int transient)
+{
+	char buf[32];
+	if (obj->type == JS_CARRAY && obj->u.a.simple && k >= 0 && k <= obj->u.a.flat_length) {
+		jsR_setarrayindex(J, obj, k, stackidx(J, -1));
+	} else {
+		jsR_setproperty(J, obj, js_itoa(buf, k), transient);
+	}
+}
+
 static void jsR_defproperty(js_State *J, js_Object *obj, const char *name,
 	int atts, js_Value *value, js_Object *getter, js_Object *setter,
 	int throw)
@@ -667,6 +820,8 @@ static void jsR_defproperty(js_State *J, js_Object *obj, const char *name,
 	if (obj->type == JS_CARRAY) {
 		if (!strcmp(name, "length"))
 			goto readonly;
+		if (obj->u.a.simple)
+			jsR_unflattenarray(J, obj);
 	}
 
 	else if (obj->type == JS_CSTRING) {
@@ -728,6 +883,8 @@ static int jsR_delproperty(js_State *J, js_Object *obj, const char *name)
 	if (obj->type == JS_CARRAY) {
 		if (!strcmp(name, "length"))
 			goto dontconf;
+		if (obj->u.a.simple)
+			jsR_unflattenarray(J, obj);
 	}
 
 	else if (obj->type == JS_CSTRING) {
@@ -765,6 +922,16 @@ dontconf:
 	return 0;
 }
 
+static void jsR_delindex(js_State *J, js_Object *obj, int k)
+{
+	char buf[32];
+	/* Allow deleting last element of a simple array without unflattening */
+	if (obj->type == JS_CARRAY && obj->u.a.simple && k == obj->u.a.flat_length - 1)
+		obj->u.a.flat_length = k;
+	else
+		jsR_delproperty(J, obj, js_itoa(buf, k));
+}
+
 /* Registry, global and object property accessors */
 
 const char *js_ref(js_State *J)
@@ -772,7 +939,7 @@ const char *js_ref(js_State *J)
 	js_Value *v = stackidx(J, -1);
 	const char *s;
 	char buf[32];
-	switch (v->type) {
+	switch (v->t.type) {
 	case JS_TUNDEFINED: s = "_Undefined"; break;
 	case JS_TNULL: s = "_Null"; break;
 	case JS_TBOOLEAN:
@@ -867,6 +1034,27 @@ int js_hasproperty(js_State *J, int idx, const char *name)
 	return jsR_hasproperty(J, js_toobject(J, idx), name);
 }
 
+void js_getindex(js_State *J, int idx, int i)
+{
+	jsR_getindex(J, js_toobject(J, idx), i);
+}
+
+int js_hasindex(js_State *J, int idx, int i)
+{
+	return jsR_hasindex(J, js_toobject(J, idx), i);
+}
+
+void js_setindex(js_State *J, int idx, int i)
+{
+	jsR_setindex(J, js_toobject(J, idx), i, !js_isobject(J, idx));
+	js_pop(J, 1);
+}
+
+void js_delindex(js_State *J, int idx, int i)
+{
+	jsR_delindex(J, js_toobject(J, idx), i);
+}
+
 /* Iterator */
 
 void js_pushiterator(js_State *J, int idx, int own)
@@ -1077,16 +1265,23 @@ static void jsR_callscript(js_State *J, int n, js_Function *F, js_Environment *s
 
 static void jsR_callcfunction(js_State *J, int n, int min, js_CFunction F)
 {
+	int save_top;
 	int i;
 	js_Value v;
 
 	for (i = n; i < min; ++i)
 		js_pushundefined(J);
 
+	save_top = TOP;
 	F(J);
-	v = *stackidx(J, -1);
-	TOP = --BOT; /* clear stack */
-	js_pushvalue(J, v);
+	if (TOP > save_top) {
+		v = *stackidx(J, -1);
+		TOP = --BOT; /* clear stack */
+		js_pushvalue(J, v);
+	} else {
+		TOP = --BOT; /* clear stack */
+		js_pushundefined(J);
+	}
 }
 
 static void jsR_pushtrace(js_State *J, const char *name, const char *file, int line)
@@ -1104,6 +1299,9 @@ void js_call(js_State *J, int n)
 	js_Object *obj;
 	int savebot;
 
+	if (n < 0)
+		js_rangeerror(J, "number of arguments cannot be negative");
+
 	if (!js_iscallable(J, -n-2))
 		js_typeerror(J, "%s is not callable", js_typeof(J, -n-2));
 
@@ -1173,13 +1371,18 @@ void js_construct(js_State *J, int n)
 	if (n > 0)
 		js_rot(J, n + 1);
 
+	/* and save a copy to return */
+	js_pushobject(J, newobj);
+	js_rot(J, n + 3);
+
 	/* call the function */
 	js_call(J, n);
 
 	/* if result is not an object, return the original object we created */
 	if (!js_isobject(J, -1)) {
 		js_pop(J, 1);
-		js_pushobject(J, newobj);
+	} else {
+		js_rot2pop1(J);
 	}
 }
 
@@ -1226,7 +1429,7 @@ int js_pcall(js_State *J, int n)
 void *js_savetrypc(js_State *J, js_Instruction *pc)
 {
 	if (J->trytop == JS_TRYLIMIT)
-		js_error(J, "try: exception stack overflow");
+		js_trystackoverflow(J);
 	J->trybuf[J->trytop].E = J->E;
 	J->trybuf[J->trytop].envtop = J->envtop;
 	J->trybuf[J->trytop].tracetop = J->tracetop;
@@ -1240,7 +1443,7 @@ void *js_savetrypc(js_State *J, js_Instruction *pc)
 void *js_savetry(js_State *J)
 {
 	if (J->trytop == JS_TRYLIMIT)
-		js_error(J, "try: exception stack overflow");
+		js_trystackoverflow(J);
 	J->trybuf[J->trytop].E = J->E;
 	J->trybuf[J->trytop].envtop = J->envtop;
 	J->trybuf[J->trytop].tracetop = J->tracetop;
@@ -1279,28 +1482,49 @@ void js_throw(js_State *J)
 
 /* Main interpreter loop */
 
-static void jsR_dumpstack(js_State *J)
+static void js_dumpvalue(js_State *J, js_Value v)
 {
-	int i;
-	printf("stack {\n");
-	for (i = 0; i < TOP; ++i) {
-		putchar(i == BOT ? '>' : ' ');
-		printf("%4d: ", i);
-		js_dumpvalue(J, STACK[i]);
-		putchar('\n');
+	switch (v.t.type) {
+	case JS_TUNDEFINED: printf("undefined"); break;
+	case JS_TNULL: printf("null"); break;
+	case JS_TBOOLEAN: printf(v.u.boolean ? "true" : "false"); break;
+	case JS_TNUMBER: printf("%.9g", v.u.number); break;
+	case JS_TSHRSTR: printf("'%s'", v.u.shrstr); break;
+	case JS_TLITSTR: printf("'%s'", v.u.litstr); break;
+	case JS_TMEMSTR: printf("'%s'", v.u.memstr->p); break;
+	case JS_TOBJECT:
+		if (v.u.object == J->G) {
+			printf("[Global]");
+			break;
+		}
+		switch (v.u.object->type) {
+		case JS_COBJECT: printf("[Object %p]", (void*)v.u.object); break;
+		case JS_CARRAY: printf("[Array %p]", (void*)v.u.object); break;
+		case JS_CFUNCTION:
+			printf("[Function %p, %s, %s:%d]",
+				(void*)v.u.object,
+				v.u.object->u.f.function->name,
+				v.u.object->u.f.function->filename,
+				v.u.object->u.f.function->line);
+			break;
+		case JS_CSCRIPT: printf("[Script %s]", v.u.object->u.f.function->filename); break;
+		case JS_CCFUNCTION: printf("[CFunction %s]", v.u.object->u.c.name); break;
+		case JS_CBOOLEAN: printf("[Boolean %d]", v.u.object->u.boolean); break;
+		case JS_CNUMBER: printf("[Number %g]", v.u.object->u.number); break;
+		case JS_CSTRING: printf("[String'%s']", v.u.object->u.s.string); break;
+		case JS_CERROR: printf("[Error]"); break;
+		case JS_CARGUMENTS: printf("[Arguments %p]", (void*)v.u.object); break;
+		case JS_CITERATOR: printf("[Iterator %p]", (void*)v.u.object); break;
+		case JS_CUSERDATA:
+			printf("[Userdata %s %p]", v.u.object->u.user.tag, v.u.object->u.user.data);
+			break;
+		default: printf("[Object %p]", (void*)v.u.object); break;
+		}
+		break;
 	}
-	printf("}\n");
 }
 
-static void jsR_dumpenvironment(js_State *J, js_Environment *E, int d)
-{
-	printf("scope %d ", d);
-	js_dumpobject(J, E->variables);
-	if (E->outer)
-		jsR_dumpenvironment(J, E->outer, d+1);
-}
-
-void js_stacktrace(js_State *J)
+static void js_stacktrace(js_State *J)
 {
 	int n;
 	printf("stack trace:\n");
@@ -1318,18 +1542,35 @@ void js_stacktrace(js_State *J)
 	}
 }
 
+static void js_dumpstack(js_State *J)
+{
+	int i;
+	printf("stack {\n");
+	for (i = 0; i < TOP; ++i) {
+		putchar(i == BOT ? '>' : ' ');
+		printf("%4d: ", i);
+		js_dumpvalue(J, STACK[i]);
+		putchar('\n');
+	}
+	printf("}\n");
+}
+
 void js_trap(js_State *J, int pc)
 {
-	if (pc > 0) {
-		js_Function *F = STACK[BOT-1].u.object->u.f.function;
-		printf("trap at %d in function ", pc);
-		jsC_dumpfunction(J, F);
-	}
-	jsR_dumpstack(J);
-	jsR_dumpenvironment(J, J->E, 0);
+	js_dumpstack(J);
 	js_stacktrace(J);
 }
 
+static int jsR_isindex(js_State *J, int idx, int *k)
+{
+	js_Value *v = stackidx(J, idx);
+	if (v->t.type == JS_TNUMBER) {
+		*k = v->u.number;
+		return *k == v->u.number && *k >= 0;
+	}
+	return 0;
+}
+
 static void jsR_run(js_State *J, js_Function *F)
 {
 	js_Function **FT = F->funtab;
@@ -1357,6 +1598,12 @@ static void jsR_run(js_State *J, js_Function *F)
 	pc += sizeof(str) / sizeof(*pc)
 
 	while (1) {
+		if (J->runlimit > 0) {
+			if (J->runlimit == 1)
+				js_runlimit(J);
+			--J->runlimit;
+		}
+
 		if (J->gccounter > J->gcthresh)
 			js_gc(J, 0);
 
@@ -1476,6 +1723,9 @@ static void jsR_run(js_State *J, js_Function *F)
 			js_pushboolean(J, b);
 			break;
 
+		case OP_SKIPARRAY:
+			js_setlength(J, -1, js_getlength(J, -1) + 1);
+			break;
 		case OP_INITARRAY:
 			js_setindex(J, -2, js_getlength(J, -2));
 			break;
@@ -1502,9 +1752,14 @@ static void jsR_run(js_State *J, js_Function *F)
 			break;
 
 		case OP_GETPROP:
-			str = js_tostring(J, -1);
-			obj = js_toobject(J, -2);
-			jsR_getproperty(J, obj, str);
+			if (jsR_isindex(J, -1, &ix)) {
+				obj = js_toobject(J, -2);
+				jsR_getindex(J, obj, ix);
+			} else {
+				str = js_tostring(J, -1);
+				obj = js_toobject(J, -2);
+				jsR_getproperty(J, obj, str);
+			}
 			js_rot3pop2(J);
 			break;
 
@@ -1516,10 +1771,16 @@ static void jsR_run(js_State *J, js_Function *F)
 			break;
 
 		case OP_SETPROP:
-			str = js_tostring(J, -2);
-			obj = js_toobject(J, -3);
-			transient = !js_isobject(J, -3);
-			jsR_setproperty(J, obj, str, transient);
+			if (jsR_isindex(J, -2, &ix)) {
+				obj = js_toobject(J, -3);
+				transient = !js_isobject(J, -3);
+				jsR_setindex(J, obj, ix, transient);
+			} else {
+				str = js_tostring(J, -2);
+				obj = js_toobject(J, -3);
+				transient = !js_isobject(J, -3);
+				jsR_setproperty(J, obj, str, transient);
+			}
 			js_rot3pop2(J);
 			break;
 
@@ -1560,7 +1821,7 @@ static void jsR_run(js_State *J, js_Function *F)
 				obj = js_toobject(J, -1);
 				str = jsV_nextiterator(J, obj);
 				if (str) {
-					js_pushliteral(J, str);
+					js_pushstring(J, str);
 					js_pushboolean(J, 1);
 				} else {
 					js_pop(J, 1);
